/*
 * MainFrame.java
 *
 * Created on Jul 14, 2011, 5:31:45 PM
 */
package codemaptournament;

import java.awt.Cursor;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.DefaultComboBoxModel;
import javax.swing.DefaultListModel;
import javax.swing.JOptionPane;
import javax.swing.JTable;
import javax.swing.RowFilter;
import javax.swing.RowSorter;
import javax.swing.SortOrder;
import javax.swing.SwingConstants;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableModel;
import javax.swing.table.TableRowSorter;
import org.pojava.datetime.DateTime;

/**
 * Holds gui stuff for Code Map Tournament Scorer
 * also a bunch of application logic because I'm lazy like that.
 * @author Paul
 */
public class MainFrame extends javax.swing.JFrame {
    private final ScoreFetcher scoreFetcher;
    /** list of last missions for which we have data. changed only by doFetchScores */
    private final List<Mission> cachedMissions;
    private Map<String, Map<Integer, Record>> scoresByGameByUser;
    private List<HighScoreEntry> highScores;
    private Map<String, HighScoreEntry> highScoresByUser;
    private Map<String, ScoreTotals> totalScoresByUser;
    private TableRowSorter tablePlayersRowSorter;
    private final RowFilter tablePlayersRowFilter;
    private final MissionTableModel missionTableModel;
    private DateTime scoreTimestamp;
    private String scoreTimestampStr;

    /** Creates new form MainFrame */
    public MainFrame() {
        cachedMissions = new ArrayList<Mission>();
        missionTableModel = new MissionTableModel();
        initComponents();
        // set some gui stuff I didn't feel like putting in initComponents
        // listener that syncs the selected tournament with the tournament missions table
        lAutoTournaments.addListSelectionListener(new ListSelectionListener() {
            public void valueChanged(ListSelectionEvent e) {
                if (e.getValueIsAdjusting()) return;
                DisplayAndValue dav = (DisplayAndValue) lAutoTournaments.getSelectedValue();
                SimpleMissionTableModel smtm = (SimpleMissionTableModel) tableTournamentMissions.getModel();
                if (dav == null) smtm.setMissionList(null);
                else {
                    List<Mission> missions = (List<Mission>) dav.value;
                    smtm.setMissionList(missions);
                }
            }
        });
        // row filter to include only those who have played all games in the Players Results table
        // enabled by calling tablePlayersRowSorter.setRowFilter(null | this)
        tablePlayersRowFilter = new RowFilter() {
            @Override
            public boolean include(Entry entry) {
                final String player = (String) entry.getValue(0);
                HighScoreEntry hse = highScoresByUser.get(player);
                return hse.playedAllGames;
            }
        };

        scoreFetcher = new ScoreFetcher();
        scoreFetcher.setSort(ScoreFetcher.SORT_SCORES);
        getRootPane().setDefaultButton(bFetchScores);

        // load missions from memory right now
        SaveLoad.load(missionTableModel);

        syncRepeatedTypesWarning();
        missionTableModel.addRepeatedMissionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                // after the data in the model has changed, it will recalculate
                // which/whether any mission have repeating size/complexity
                // types. if so, we might need to show/hide the warning
                // above the table.
                syncRepeatedTypesWarning();
            }
        });

        // set up to save missions right before the window closes
        addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent e) {
                SaveLoad.save(missionTableModel);
            }

        });
    }

    /**
     * Look at forum posts to get recent tournaments + maps for each.
     * Modifies the lAutoTournaments list.
     */
    private void doAutoImportTournaments() {
        final Cursor oldCursor = getCursor();
        try {
            setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
            
            // get the model of lists of tournaments
            final DefaultListModel lm = (DefaultListModel) lAutoTournaments.getModel();
            lm.clear();
            
            // Fetch the missions
            AutoWeeklyMissionFetcher fetcher = new AutoWeeklyMissionFetcher();
            Map<WeekInfo, List<Mission>> weeklyMissionData = fetcher.getWeeklyMissionData();
            
            // Store the missions into the list model
            for (Map.Entry<WeekInfo, List<Mission>> entry : weeklyMissionData.entrySet()) {
                WeekInfo weekInfo = entry.getKey();
                List<Mission> missions = entry.getValue();
                // make label with format: "(Topic) closing (Date string)   (maybe [Open])"
                StringBuilder label = new StringBuilder();
                label.append(weekInfo.topic);
                label.append(" closing ");
                label.append(weekInfo.closing);
                if (weekInfo.isOpen())
                    label.append("  [Open]");
                // store as label -> list of missions using DisplayAndValue
                DisplayAndValue week = new DisplayAndValue(label.toString(), missions);
                lm.addElement(week);
            }
            
            // if we got any missions, set the first one as default. otherwise, complain.
            if (lm.size() > 0) {
                lAutoTournaments.setSelectedIndex(0);
                bLoadTournament.setEnabled(true);
            } else {
                bLoadTournament.setEnabled(false);
                JOptionPane.showMessageDialog(this, "No missions found.", "Failure!", JOptionPane.ERROR_MESSAGE);
            }
        } catch (Exception ex) {
            bLoadTournament.setEnabled(false);
            Logger.getLogger(MainFrame.class.getName()).log(Level.WARNING, "auto fetch tournament failure", ex);
            JOptionPane.showMessageDialog(this, ex, "Something went wrong!", JOptionPane.ERROR_MESSAGE);
        } finally {
            setCursor(oldCursor);
        }
    }


    /**
     * Create forum list of high scores for the tournament and copy to clipboard.
     */
    private void doCopyHighscores() {
        boolean boldTop3 = cbHSBoldColors.isSelected();
        // generate high score list
        final String report = Reports.makeForumHighScoreList(highScores, scoreTimestampStr, boldTop3);
        // export to clipboard and announce it
        ClipboardUtil.setClipboardContents(report);
        JOptionPane.showMessageDialog(this, "High score list has been copied to the clipboard.\nPaste in a forum post.", "Success", JOptionPane.INFORMATION_MESSAGE);
    }

    /**
     * Create forum list of missions in the table and copy to clipboard.
     * @param extendedTable whether to make the extended table (with gravity/build/exp data)
     */
    private void doCopyMissionList(boolean extendedTable, boolean includeStandardHeader) {
        final List<Mission> missions = missionTableModel.getMissionList();

        // make sure there are no repeated mission types
        // the model will already be up to date since it listens to any
        // of its own changes
        final boolean repeatedTypes = missionTableModel.anyRowsBad();
        if (repeatedTypes) {
            // warn the user, but still allow him to export if he really wants to
            final String repeated_type_message = "The rules state that no two maps in a tournament \ncan have the same combination of size/complexity.\n\nContinue anyway?";
            int result = JOptionPane.showConfirmDialog(
                    this,
                    repeated_type_message,
                    "Repeated Size and Complexity",
                    JOptionPane.YES_NO_OPTION,
                    JOptionPane.WARNING_MESSAGE);
            if (result != JOptionPane.YES_OPTION) return;
        }

        // generate mission list
        StringBuilder report = new StringBuilder();
        if (includeStandardHeader) report.append(Reports.TOURNAMENT_PREAMBLE);
        if (extendedTable) report.append(Reports.makeForumTournamentListExtended(missions));
        else               report.append(Reports.makeForumTournamentList(missions));
        
        // export to clipboard and announce it
        ClipboardUtil.setClipboardContents(report.toString());
        String message = "Mission List table has been copied to the clipboard.\nPaste in a forum post.";
        if (extendedTable) message = message + "\n\nSET GRAVITY, BUILD, AND EXPERIMENTALS INFO BY HAND!";
        JOptionPane.showMessageDialog(this, message, "Success", JOptionPane.INFORMATION_MESSAGE);
    }

    /**
     * Copy the current table of player results to a BBCode table.
     * This table has scores for each map (with top scores indicated) as well as
     * overall high scores.
     */
    private void doCopyResults() {
        // set up the table to sort by total score descending
        final int numMissions = cachedMissions.size();
        final int COL_TOTAL_SCORE = numMissions+1;
        List<RowSorter.SortKey> sortKeys = new ArrayList<RowSorter.SortKey>();
        sortKeys.add(new RowSorter.SortKey(COL_TOTAL_SCORE, SortOrder.DESCENDING));
        final RowSorter<? extends TableModel> rowSorter = tablePlayerResults.getRowSorter();
        List<? extends RowSorter.SortKey> oldSortKeys = rowSorter.getSortKeys();
        rowSorter.setSortKeys(sortKeys);
        // get the report
        boolean colorMapTop3 = cbResultsColorTop3.isSelected();
        boolean boldTop3 = cbResultsBoldColors.isSelected();
        final String report = Reports.makeForumDetailedResults(
                tablePlayerResults, cachedMissions, scoreTimestampStr,
                colorMapTop3, boldTop3
        );
        // restore old sort order
        rowSorter.setSortKeys(oldSortKeys);
        // export to clipboard and announce it
        ClipboardUtil.setClipboardContents(report);
        JOptionPane.showMessageDialog(this, "Detailed mission result table has been copied to the clipboard.\nPaste in a forum post.", "Success", JOptionPane.INFORMATION_MESSAGE);
    }

    /**
     * Get scores for the missions the user has entered and display the results,
     * totals, etc in the various tabs.
     * Shows wait cursor. Shows dialog on exception.
     */
    private void doFetchScores() {
        // this is where the magic happens
        final Cursor oldCursor = getCursor();
        try {
            setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

            // read in GUI stuff to initialize score fetcher
            scoreFetcher.setGroupfilter(tfGroupFilter.getText());
            scoreFetcher.setTimefilter((Integer) ((DisplayAndValue)cbTimeFilter.getSelectedItem()).value);

            // clear GUI stuff in case no scores are fetched
            taMissionScores.setText("");
            taRankings.setText("");
            bCopyHighScores.setEnabled(false);
            bCopyResults.setEnabled(false);
            // turn off row filter since the model's state isn't valid until everything has been set
            tablePlayersRowSorter.setRowFilter(null);

            readMissionTable();
            List<List<Record>> recordsByMission = fetchMissionScoreData();
            // TODO maybe filter out banned users?
            tallyScores(recordsByMission);
            updateMissionScoresTab(recordsByMission);
            updatePlayerResultsTab();
            updateHighScoresTab();

            // enable gui stuff
            bCopyHighScores.setEnabled(true);
            bCopyResults.setEnabled(true);
            syncPlayerRowFilter();
        } catch (Exception ex) {
            Logger.getLogger(MainFrame.class.getName()).log(Level.WARNING, "fetch score failure", ex);
            JOptionPane.showMessageDialog(this, ex, "Something went wrong!", JOptionPane.ERROR_MESSAGE);
        } finally {
            setCursor(oldCursor);
        }
    }


    /** transfer the mission data in the table into cachedMissions */
    private void readMissionTable() {
        cachedMissions.clear();
        cachedMissions.addAll(missionTableModel.getMissionList());
    }

    /** Fetch data for each mission info in cachedMissions
     * @return list records for each mission
     * @throws Exception if fetch fails
     */
    private List<List<Record>> fetchMissionScoreData() throws Exception {
        final int missionCount = cachedMissions.size();
        // fetch the score records
        final List<List<Record>> recordsByMission = new ArrayList<List<Record>>(missionCount);
        for (Mission mission : cachedMissions) {
            List<Record> records = scoreFetcher.fetchScoreRecords(mission.gameUID);
            recordsByMission.add(records);
        }
        // record the current time as scoreTimestampStr
        scoreTimestamp = new DateTime();
        scoreTimestampStr = scoreTimestamp.toString(ForumTime.FORUM_TZ);
        return recordsByMission;
    }

    /** Tally up the scores given in the list of records for each mission.
     *  Also sets scoresByGameByUser, totalScoresByUser, highScores, highScoresByUser
     * @param recordsByMission
     */
    private void tallyScores(final List<List<Record>> recordsByMission) {
        final int missionCount = recordsByMission.size();
        // add up the scores using a ResultRanker
        final ResultRanker ranker = new ResultRanker(missionCount);
        for (int gameNumber = 0; gameNumber < missionCount; gameNumber++) {
            for (Record record : recordsByMission.get(gameNumber)) {
                ranker.recordScore(gameNumber, record);
            }
        }
        // get the various data structures from the ResultRanker
        scoresByGameByUser = ranker.getScoresByGameByUser();
        totalScoresByUser = ranker.getTotalsByUser();
        highScores = ranker.getHighScoreList();
        // make lookup map of user to high score from the high score list
        highScoresByUser = new TreeMap<String, HighScoreEntry>();
        for (HighScoreEntry hse : highScores) {
            highScoresByUser.put(hse.user, hse);
        }
    }

    /** Update data for Tab 1 - Mission Scores */
    private void updateMissionScoresTab(final List<List<Record>> recordsByMission) {
        final int missionCount = cachedMissions.size();
        final int userCount = totalScoresByUser.size();
        taMissionScores.append("Player\tScore\tTime\t# Plays\n\n");
        // fill in the text area with all the scores for each mission
        int scoreCount = 0;
        for (int i = 0; i < missionCount; i++) {
            Mission m = cachedMissions.get(i);
            List<Record> records = recordsByMission.get(i);
            taMissionScores.append("--- Mission " + (i + 1) + ": " + m.code + " (" + records.size() + " scores) ---\n");
            for (Record record : records) {
                taMissionScores.append(record.user);
                taMissionScores.append("\t" + record.score);
                taMissionScores.append("\t" + record.time.toString());
                taMissionScores.append("\t" + record.plays);
                taMissionScores.append("\n");
                scoreCount++;
            }
        }
        // announce some stats at top
        taMissionScores.insert(scoreCount + " scores fetched for " + missionCount + " missions and " + userCount + " players at " + scoreTimestampStr + " (forum time).\n", 0);
        taMissionScores.setCaretPosition(0); // keep the cursor at the beginning
    }

    /** Update data for Tab 2 - Player Results */
    private void updatePlayerResultsTab() {
        final int missionCount = cachedMissions.size();
        final int userCount = totalScoresByUser.size();
        // go through each player and record their data in the tablePlayerResults table
        final DefaultTableModel mPlayerResults = (DefaultTableModel) tablePlayerResults.getModel();
        mPlayerResults.setRowCount(userCount);
        int row = 0;
        for (Map.Entry<String, Map<Integer, Record>> userEntry : scoresByGameByUser.entrySet()) {
            String user = userEntry.getKey();
            Map<Integer, Record> scoresByGame = userEntry.getValue();
            ScoreTotals totals = totalScoresByUser.get(user);
            int col = 0;
            boolean playedAllGames = true;
            mPlayerResults.setValueAt(user, row, col++);
            for (int i = 0; i < missionCount; i++) {
                Record record = scoresByGame.get(i);
                Integer score = (record == null) ? null : record.score; // null not 0 so it is blank when no score is available
                mPlayerResults.setValueAt(score, row, col++);
                if (record == null) {
                    playedAllGames = false;
                }
            }
            mPlayerResults.setValueAt(totals.getTotalScore(), row, col++);
            mPlayerResults.setValueAt(playedAllGames ? totals.getTotalTimeAsFrames() : null, row, col++);
            row++;
        }
    }

    /** Update data for Tab 3 - Rankings */
    // TODO maybe use a table for high score list?
    private void updateHighScoresTab() {
        final int missionCount = cachedMissions.size();
        taRankings.setText("Current High Scores:\n");
        int rank = 1;
        for (HighScoreEntry entry : highScores) {
            taRankings.append(rank + ": " + entry.user + "\t" + entry.totalScore);
            if (!entry.playedAllGames) {
                taRankings.append("*");
            }
            taRankings.append("\n");
            rank++;
        }
        taRankings.append("\n* Did not finish all " + missionCount + " missions.");
        taRankings.setCaretPosition(0); // keep the cursor at the beginning
    }

    // was called by a dump table button during development
    private void dumpTableSizes(JTable table) {
        TableColumnModel columnModel = table.getColumnModel();
        int count = columnModel.getColumnCount();
        for (int i = 0; i < count; i++) {
            System.out.println("width(" + i + ")=" + columnModel.getColumn(i).getWidth());
        }
        System.out.println("total width = " + columnModel.getTotalColumnWidth());
        System.out.println("table height = " + table.getHeight());
    }

    /** get a DefaultTableCellRenderer with right horizontal justification */
    private TableCellRenderer getRightJustifiedTableRenderer() {
        DefaultTableCellRenderer rend = new DefaultTableCellRenderer();
        rend.setHorizontalAlignment(SwingConstants.RIGHT);
        return rend;
    }

    /** set the row filter based on the selection status of the check button */
    private void syncPlayerRowFilter() {
        if (cbHideIncompletePlayers.isSelected()) {
            tablePlayersRowSorter.setRowFilter(tablePlayersRowFilter);
        } else {
            tablePlayersRowSorter.setRowFilter(null);
        }
    }

    /** hide and show the message above the table about repeated types */
    private void syncRepeatedTypesWarning() {
        lMissionMessage.setVisible(missionTableModel.anyRowsBad());
    }


    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        javax.swing.JLabel lMissions = new javax.swing.JLabel();
        bFetchScores = new javax.swing.JButton();
        spMissions = new javax.swing.JScrollPane();
        tableMissions = new javax.swing.JTable();
        resultPanel = new javax.swing.JTabbedPane();
        spMissionScores = new javax.swing.JScrollPane();
        taMissionScores = new javax.swing.JTextArea();
        pPlayerResults = new javax.swing.JPanel();
        spPlayerResults = new javax.swing.JScrollPane();
        tablePlayerResults = new javax.swing.JTable();
        cbHideIncompletePlayers = new javax.swing.JCheckBox();
        spRankings = new javax.swing.JScrollPane();
        taRankings = new javax.swing.JTextArea();
        exportPanel = new javax.swing.JPanel();
        javax.swing.JLabel lExportHeader = new javax.swing.JLabel();
        bCopyMissionList = new javax.swing.JButton();
        bCopyHighScores = new javax.swing.JButton();
        cbMissionListExtented = new javax.swing.JCheckBox();
        cbMissionListHeader = new javax.swing.JCheckBox();
        bCopyResults = new javax.swing.JButton();
        cbResultsColorTop3 = new javax.swing.JCheckBox();
        cbResultsBoldColors = new javax.swing.JCheckBox();
        cbHSBoldColors = new javax.swing.JCheckBox();
        importPanel = new javax.swing.JPanel();
        javax.swing.JLabel lImportHeader = new javax.swing.JLabel();
        bGetTournaments = new javax.swing.JButton();
        bLoadTournament = new javax.swing.JButton();
        javax.swing.JLabel jLabel7 = new javax.swing.JLabel();
        spAutoTournaments = new javax.swing.JScrollPane();
        lAutoTournaments = new javax.swing.JList();
        javax.swing.JLabel jLabel8 = new javax.swing.JLabel();
        spTournamentMissions = new javax.swing.JScrollPane();
        tableTournamentMissions = new javax.swing.JTable();
        javax.swing.JLabel lGroupFilter = new javax.swing.JLabel();
        tfGroupFilter = new javax.swing.JTextField();
        javax.swing.JLabel lTimeFilter = new javax.swing.JLabel();
        cbTimeFilter = new javax.swing.JComboBox();
        lMissionMessage = new javax.swing.JLabel();

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        setTitle("Code Map Tournament Scorer");
        setMinimumSize(new java.awt.Dimension(600, 480));

        lMissions.setLabelFor(tableMissions);
        lMissions.setText("Tournament Missions:");
        lMissions.setToolTipText("<html>Info for all 5 missions in the tournament.<br>\nYou can enter by hand or drag and drop from a table.</html>");

        bFetchScores.setText("Fetch Scores");
        bFetchScores.setToolTipText("<html>Get scores from knucklecracker.com for the 5 listed maps.<br>If there's a problem, check your internet connection.</html>");
        bFetchScores.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                bFetchScoresActionPerformed(evt);
            }
        });

        tableMissions.setModel(missionTableModel);
        tableMissions.setColumnSelectionAllowed(true);
        tableMissions.setDragEnabled(true);
        tableMissions.setDropMode(javax.swing.DropMode.ON);
        tableMissions.setFillsViewportHeight(true);
        tableMissions.setPreferredSize(new java.awt.Dimension(435, 90));
        tableMissions.setRowHeight(18);
        tableMissions.setSelectionMode(javax.swing.ListSelectionModel.SINGLE_INTERVAL_SELECTION);
        tableMissions.putClientProperty("terminateEditOnFocusLost", Boolean.TRUE);
        MissionTableModel.setUpMissionTableColumns(tableMissions);
        MissionTableModel.setUpDropHandler(tableMissions);
        spMissions.setViewportView(tableMissions);

        resultPanel.setMinimumSize(new java.awt.Dimension(102, 87));

        taMissionScores.setEditable(false);
        taMissionScores.setLineWrap(true);
        taMissionScores.setTabSize(15);
        taMissionScores.setText("Code Map Tournament Scorer\nBy mopa42\nv0.6\n\nFill in the details for the 5 code missions, then press the button to fetch scores.\nThe scores are displayed three ways in the first three tabs:\n- Scores by mission by rank (like the normal score viewer)\n- Scores by player for each mission\n- Overall rankings\n\nIn the fourth tab you can copy forum-formatted tables of some of the data.\nIn the fifth tab you can get the list of this week's missions automatically by checking the forum.\n\nFor tournaments all maps should have different size/complexity types. The program will warn you if you have any repeats.\n\nEnjoy!\nFeedback is welcome.");
        taMissionScores.setWrapStyleWord(true);
        taMissionScores.setDragEnabled(true);
        spMissionScores.setViewportView(taMissionScores);

        resultPanel.addTab("Mission Scores", null, spMissionScores, "Scores and times for each map (fresh from the server).");

        tablePlayerResults.setModel(new javax.swing.table.DefaultTableModel(
            new Object [][] {

            },
            new String [] {
                "Name", "Score 1", "Score 2", "Score 3", "Score 4", "Score 5", "Total Score", "Total Time"
            }
        ) {
            Class[] types = new Class [] {
                java.lang.String.class, java.lang.Integer.class, java.lang.Integer.class, java.lang.Integer.class, java.lang.Integer.class, java.lang.Integer.class, java.lang.Integer.class, java.lang.Object.class
            };
            boolean[] canEdit = new boolean [] {
                false, false, false, false, false, false, false, false
            };

            public Class getColumnClass(int columnIndex) {
                return types [columnIndex];
            }

            public boolean isCellEditable(int rowIndex, int columnIndex) {
                return canEdit [columnIndex];
            }
        });
        // SUPER UGLY HACK to change the class of the total times column to Frames so the sorter works
        // I just copy pasted from the line above, so TODO change this if I change the player results table
        tablePlayerResults.setModel(new javax.swing.table.DefaultTableModel(
            new Object [][] {

            },
            new String [] {
                "Name", "Score 1", "Score 2", "Score 3", "Score 4", "Score 5", "Total Score", "Total Time"
            }
        ) {
            Class[] types = new Class [] {
                java.lang.String.class, java.lang.Integer.class, java.lang.Integer.class, java.lang.Integer.class, java.lang.Integer.class, java.lang.Integer.class, java.lang.Integer.class, Frames.class
            };
            boolean[] canEdit = new boolean [] {
                false, false, false, false, false, false, false, false
            };

            public Class getColumnClass(int columnIndex) {
                return types [columnIndex];
            }

            public boolean isCellEditable(int rowIndex, int columnIndex) {
                return canEdit [columnIndex];
            }
        });
        tablePlayerResults.setDragEnabled(true);
        tablePlayerResults.setFillsViewportHeight(true);
        TableRowSorter rowSorter = new TableRowSorter(tablePlayerResults.getModel());
        // sort names case insensitively
        rowSorter.setComparator(0, String.CASE_INSENSITIVE_ORDER);
        // sort total times with null times higher than nonnull
        //rowSorter.setComparator(7, new NullComparator(true));
        // THE SORTER HANDLES NULLS ITSELF! WE CAN'T CHANGE IT!
        tablePlayerResults.setRowSorter(rowSorter);
        tablePlayersRowSorter = rowSorter;
        spPlayerResults.setViewportView(tablePlayerResults);
        tablePlayerResults.getColumnModel().getColumn(0).setPreferredWidth(100);
        tablePlayerResults.getColumnModel().getColumn(1).setPreferredWidth(60);
        tablePlayerResults.getColumnModel().getColumn(2).setPreferredWidth(60);
        tablePlayerResults.getColumnModel().getColumn(3).setPreferredWidth(60);
        tablePlayerResults.getColumnModel().getColumn(4).setPreferredWidth(60);
        tablePlayerResults.getColumnModel().getColumn(5).setPreferredWidth(60);
        tablePlayerResults.getColumnModel().getColumn(6).setPreferredWidth(80);
        tablePlayerResults.getColumnModel().getColumn(7).setPreferredWidth(80);
        tablePlayerResults.getColumnModel().getColumn(7).setCellRenderer(getRightJustifiedTableRenderer());

        cbHideIncompletePlayers.setText("Hide players who have not completed all missions");
        cbHideIncompletePlayers.setToolTipText("Show only the players who have completed all the missions.");
        cbHideIncompletePlayers.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                cbHideIncompletePlayersActionPerformed(evt);
            }
        });

        javax.swing.GroupLayout pPlayerResultsLayout = new javax.swing.GroupLayout(pPlayerResults);
        pPlayerResults.setLayout(pPlayerResultsLayout);
        pPlayerResultsLayout.setHorizontalGroup(
            pPlayerResultsLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(pPlayerResultsLayout.createSequentialGroup()
                .addComponent(cbHideIncompletePlayers)
                .addContainerGap())
            .addComponent(spPlayerResults, javax.swing.GroupLayout.DEFAULT_SIZE, 555, Short.MAX_VALUE)
        );
        pPlayerResultsLayout.setVerticalGroup(
            pPlayerResultsLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, pPlayerResultsLayout.createSequentialGroup()
                .addComponent(spPlayerResults, javax.swing.GroupLayout.DEFAULT_SIZE, 282, Short.MAX_VALUE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                .addComponent(cbHideIncompletePlayers)
                .addContainerGap())
        );

        resultPanel.addTab("Results Per Player", null, pPlayerResults, "Scores and totals for each player. Click on a column to enable sorting.");

        taRankings.setColumns(20);
        taRankings.setRows(5);
        taRankings.setTabSize(15);
        taRankings.setDragEnabled(true);
        spRankings.setViewportView(taRankings);

        resultPanel.addTab("Rankings", null, spRankings, "High score chart.");

        lExportHeader.setText("Use these buttons to copy to the clipboard text in BB-code format (for the forum).");

        bCopyMissionList.setText("Copy Missions List as BB Table");
        bCopyMissionList.setToolTipText("Export list of missions for starting a tournament thread.");
        bCopyMissionList.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                bCopyMissionListActionPerformed(evt);
            }
        });

        bCopyHighScores.setText("Copy High Score list as BB Table");
        bCopyHighScores.setToolTipText("Export the high scores for this tournament.");
        bCopyHighScores.setEnabled(false);
        bCopyHighScores.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                bCopyHighScoresActionPerformed(evt);
            }
        });

        cbMissionListExtented.setText("Extended Table");
        cbMissionListExtented.setToolTipText("<html>Also list Gravity, Build, and Experimentals columns in table.\n<br>You must use the game to fill in these values by hand since I can't compute them.\n</html>");

        cbMissionListHeader.setText("Include Header");
        cbMissionListHeader.setToolTipText("Add the standard tournament introduction before the table of missions.");

        bCopyResults.setText("Copy Detailed Results as BB Table");
        bCopyResults.setToolTipText("Export a high-score table with breakdown of individual maps.");
        bCopyResults.setEnabled(false);
        bCopyResults.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                bCopyResultsActionPerformed(evt);
            }
        });

        cbResultsColorTop3.setText("Top 3 Per Map");
        cbResultsColorTop3.setToolTipText("Show gold/silver/bronze for top three scores for each map, rather than just bold top score per map.");

        cbResultsBoldColors.setText("Bold Colors");
        cbResultsBoldColors.setToolTipText("Make gold/silver/bronze stand out more by making them bold.");

        cbHSBoldColors.setText("Bold Colors");
        cbHSBoldColors.setToolTipText("Make gold/silver/bronze stand out more by making them bold.");

        javax.swing.GroupLayout exportPanelLayout = new javax.swing.GroupLayout(exportPanel);
        exportPanel.setLayout(exportPanelLayout);
        exportPanelLayout.setHorizontalGroup(
            exportPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(exportPanelLayout.createSequentialGroup()
                .addGroup(exportPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(exportPanelLayout.createSequentialGroup()
                        .addGap(10, 10, 10)
                        .addComponent(lExportHeader, javax.swing.GroupLayout.DEFAULT_SIZE, 556, Short.MAX_VALUE))
                    .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, exportPanelLayout.createSequentialGroup()
                        .addContainerGap()
                        .addGroup(exportPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING)
                            .addComponent(bCopyMissionList, javax.swing.GroupLayout.DEFAULT_SIZE, 195, Short.MAX_VALUE)
                            .addComponent(bCopyHighScores, javax.swing.GroupLayout.DEFAULT_SIZE, 195, Short.MAX_VALUE)
                            .addComponent(bCopyResults, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
                        .addGap(18, 18, 18)
                        .addGroup(exportPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addComponent(cbHSBoldColors)
                            .addGroup(exportPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                                .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, exportPanelLayout.createSequentialGroup()
                                    .addComponent(cbMissionListHeader)
                                    .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                                    .addComponent(cbMissionListExtented)
                                    .addGap(141, 141, 141))
                                .addGroup(exportPanelLayout.createSequentialGroup()
                                    .addComponent(cbResultsBoldColors)
                                    .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                                    .addComponent(cbResultsColorTop3))))))
                .addContainerGap())
        );
        exportPanelLayout.setVerticalGroup(
            exportPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(exportPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addComponent(lExportHeader)
                .addGap(18, 18, 18)
                .addGroup(exportPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(bCopyMissionList)
                    .addComponent(cbMissionListHeader)
                    .addComponent(cbMissionListExtented))
                .addGap(18, 18, 18)
                .addGroup(exportPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(bCopyHighScores)
                    .addComponent(cbHSBoldColors))
                .addGap(18, 18, 18)
                .addGroup(exportPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(bCopyResults)
                    .addComponent(cbResultsBoldColors)
                    .addComponent(cbResultsColorTop3))
                .addGap(171, 171, 171))
        );

        resultPanel.addTab("Export", null, exportPanel, "Export data to a forum post.");

        lImportHeader.setText("Attempt to fill in details for the 5 missions by looking at posts on the forum.");

        bGetTournaments.setText("Get Tournament List");
        bGetTournaments.setToolTipText("<html>Get current weekly tournaments listed in the forum.<br>This may take a minute depending on your connection.");
        bGetTournaments.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                bGetTournamentsActionPerformed(evt);
            }
        });

        bLoadTournament.setText("Import Selected Tournament");
        bLoadTournament.setToolTipText("Copy the missions from the selected tournament below to the mission table above.");
        bLoadTournament.setEnabled(false);
        bLoadTournament.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                bLoadTournamentActionPerformed(evt);
            }
        });

        jLabel7.setLabelFor(lAutoTournaments);
        jLabel7.setText("Tournaments listed on the Code Map Discussion page:");

        lAutoTournaments.setModel(new DefaultListModel());
        lAutoTournaments.setSelectionMode(javax.swing.ListSelectionModel.SINGLE_SELECTION);
        lAutoTournaments.setToolTipText("Weekly 5-Map Tournament threads, ordered by closing date.");
        lAutoTournaments.setVisibleRowCount(5);
        spAutoTournaments.setViewportView(lAutoTournaments);

        jLabel8.setLabelFor(tableTournamentMissions);
        jLabel8.setText("Missions from the selected tournament:");
        jLabel8.setToolTipText("If auto-import fails and this table is blank, you'll have to type in missions by hand.");

        tableTournamentMissions.setModel(new SimpleMissionTableModel());
        tableTournamentMissions.setToolTipText("The first 5 missions found in a table on this tournament's first post.");
        tableTournamentMissions.setCellSelectionEnabled(true);
        tableTournamentMissions.setDragEnabled(true);
        tableTournamentMissions.setFillsViewportHeight(true);
        tableTournamentMissions.setPreferredSize(new java.awt.Dimension(100, 80));
        tableTournamentMissions.setSelectionMode(javax.swing.ListSelectionModel.SINGLE_INTERVAL_SELECTION);
        tableTournamentMissions.setShowHorizontalLines(false);
        tableTournamentMissions.setShowVerticalLines(false);
        tableTournamentMissions.setTableHeader(null);
        tableTournamentMissions.getColumnModel().getColumn(0).setPreferredWidth(200);
        tableTournamentMissions.getColumnModel().getColumn(1).setPreferredWidth(30);
        tableTournamentMissions.getColumnModel().getColumn(2).setPreferredWidth(30);
        spTournamentMissions.setViewportView(tableTournamentMissions);

        javax.swing.GroupLayout importPanelLayout = new javax.swing.GroupLayout(importPanel);
        importPanel.setLayout(importPanelLayout);
        importPanelLayout.setHorizontalGroup(
            importPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(importPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(importPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(spAutoTournaments, javax.swing.GroupLayout.Alignment.TRAILING, javax.swing.GroupLayout.DEFAULT_SIZE, 535, Short.MAX_VALUE)
                    .addComponent(lImportHeader)
                    .addGroup(importPanelLayout.createSequentialGroup()
                        .addComponent(bGetTournaments, javax.swing.GroupLayout.DEFAULT_SIZE, 262, Short.MAX_VALUE)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                        .addComponent(bLoadTournament, javax.swing.GroupLayout.DEFAULT_SIZE, 263, Short.MAX_VALUE))
                    .addComponent(jLabel7)
                    .addComponent(spTournamentMissions, javax.swing.GroupLayout.DEFAULT_SIZE, 535, Short.MAX_VALUE)
                    .addComponent(jLabel8))
                .addContainerGap())
        );
        importPanelLayout.setVerticalGroup(
            importPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(importPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addComponent(lImportHeader)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(importPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(bGetTournaments)
                    .addComponent(bLoadTournament))
                .addGap(16, 16, 16)
                .addComponent(jLabel7)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(spAutoTournaments, javax.swing.GroupLayout.DEFAULT_SIZE, 93, Short.MAX_VALUE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                .addComponent(jLabel8)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(spTournamentMissions, javax.swing.GroupLayout.DEFAULT_SIZE, 94, Short.MAX_VALUE)
                .addContainerGap())
        );

        resultPanel.addTab("Import", null, importPanel, "Get this week's missions from the forum.");

        lGroupFilter.setLabelFor(tfGroupFilter);
        lGroupFilter.setText("Group Filter:");

        tfGroupFilter.setToolTipText("(Optional) filter by group name - I don't use this but someone else might.");

        lTimeFilter.setLabelFor(cbTimeFilter);
        lTimeFilter.setText("Time Filter:");

        cbTimeFilter.setModel(new DefaultComboBoxModel(ScoreFetcher.timeFilters));
        cbTimeFilter.setSelectedItem(ScoreFetcher.timeFilters[ScoreFetcher.TIME_LAST_MONTH]);
        cbTimeFilter.setToolTipText("Timeframe for scores - choose Last Week or Last Month for a weekly tournament.");
        cbTimeFilter.setPrototypeDisplayValue(ScoreFetcher.longestTimeFilterStr);

        lMissionMessage.setFont(lMissionMessage.getFont().deriveFont((lMissionMessage.getFont().getStyle() | java.awt.Font.ITALIC) & ~java.awt.Font.BOLD));
        lMissionMessage.setHorizontalAlignment(javax.swing.SwingConstants.TRAILING);
        lMissionMessage.setText("Missions should have different size/complexity combinations!");

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING)
                    .addComponent(resultPanel, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, 560, Short.MAX_VALUE)
                    .addComponent(spMissions, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, 560, Short.MAX_VALUE)
                    .addGroup(javax.swing.GroupLayout.Alignment.LEADING, layout.createSequentialGroup()
                        .addComponent(lTimeFilter)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(cbTimeFilter, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addGap(18, 18, 18)
                        .addComponent(lGroupFilter)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(tfGroupFilter, javax.swing.GroupLayout.PREFERRED_SIZE, 102, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addGap(18, 18, 18)
                        .addComponent(bFetchScores, javax.swing.GroupLayout.DEFAULT_SIZE, 199, Short.MAX_VALUE))
                    .addGroup(javax.swing.GroupLayout.Alignment.LEADING, layout.createSequentialGroup()
                        .addComponent(lMissions)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED, 166, Short.MAX_VALUE)
                        .addComponent(lMissionMessage)))
                .addContainerGap())
        );

        layout.linkSize(javax.swing.SwingConstants.HORIZONTAL, new java.awt.Component[] {cbTimeFilter, tfGroupFilter});

        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(lMissions)
                    .addComponent(lMissionMessage))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(spMissions, javax.swing.GroupLayout.PREFERRED_SIZE, 114, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(bFetchScores)
                    .addComponent(lTimeFilter)
                    .addComponent(cbTimeFilter, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(tfGroupFilter, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(lGroupFilter))
                .addGap(18, 18, 18)
                .addComponent(resultPanel, javax.swing.GroupLayout.DEFAULT_SIZE, 347, Short.MAX_VALUE)
                .addContainerGap())
        );

        lMissionMessage.setVisible(false);

        pack();
    }// </editor-fold>//GEN-END:initComponents

    private void bFetchScoresActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bFetchScoresActionPerformed
        doFetchScores();
    }//GEN-LAST:event_bFetchScoresActionPerformed

    private void bCopyMissionListActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bCopyMissionListActionPerformed
        doCopyMissionList(cbMissionListExtented.isSelected(), cbMissionListHeader.isSelected());
    }//GEN-LAST:event_bCopyMissionListActionPerformed

    private void bCopyHighScoresActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bCopyHighScoresActionPerformed
        doCopyHighscores();
    }//GEN-LAST:event_bCopyHighScoresActionPerformed

    private void bGetTournamentsActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bGetTournamentsActionPerformed
        doAutoImportTournaments();
    }//GEN-LAST:event_bGetTournamentsActionPerformed

    private void bLoadTournamentActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bLoadTournamentActionPerformed
        DisplayAndValue dav = (DisplayAndValue) lAutoTournaments.getSelectedValue();
        List<Mission> missions = (List<Mission>) dav.value;
        int row = 0;
        for (Mission mission : missions) {
            missionTableModel.setMission(row++, mission);
        }
    }//GEN-LAST:event_bLoadTournamentActionPerformed

    private void cbHideIncompletePlayersActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_cbHideIncompletePlayersActionPerformed
        syncPlayerRowFilter();
    }//GEN-LAST:event_cbHideIncompletePlayersActionPerformed

    private void bCopyResultsActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bCopyResultsActionPerformed
        doCopyResults();
    }//GEN-LAST:event_bCopyResultsActionPerformed


    /**
     * @param args the command line arguments (ignored)
     */
    public static void main(String args[]) {
//        try {
//            UIManager.setLookAndFeel("com.sun.java.swing.plaf.nimbus.NimbusLookAndFeel");
//        } catch (Exception e) {
//        }
        java.awt.EventQueue.invokeLater(new Runnable() {

            public void run() {
                new MainFrame().setVisible(true);
            }
        });
    }
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JButton bCopyHighScores;
    private javax.swing.JButton bCopyMissionList;
    private javax.swing.JButton bCopyResults;
    private javax.swing.JButton bFetchScores;
    private javax.swing.JButton bGetTournaments;
    private javax.swing.JButton bLoadTournament;
    private javax.swing.JCheckBox cbHSBoldColors;
    private javax.swing.JCheckBox cbHideIncompletePlayers;
    private javax.swing.JCheckBox cbMissionListExtented;
    private javax.swing.JCheckBox cbMissionListHeader;
    private javax.swing.JCheckBox cbResultsBoldColors;
    private javax.swing.JCheckBox cbResultsColorTop3;
    private javax.swing.JComboBox cbTimeFilter;
    private javax.swing.JPanel exportPanel;
    private javax.swing.JPanel importPanel;
    private javax.swing.JList lAutoTournaments;
    private javax.swing.JLabel lMissionMessage;
    private javax.swing.JPanel pPlayerResults;
    private javax.swing.JTabbedPane resultPanel;
    private javax.swing.JScrollPane spAutoTournaments;
    private javax.swing.JScrollPane spMissionScores;
    private javax.swing.JScrollPane spMissions;
    private javax.swing.JScrollPane spPlayerResults;
    private javax.swing.JScrollPane spRankings;
    private javax.swing.JScrollPane spTournamentMissions;
    private javax.swing.JTextArea taMissionScores;
    private javax.swing.JTextArea taRankings;
    private javax.swing.JTable tableMissions;
    private javax.swing.JTable tablePlayerResults;
    private javax.swing.JTable tableTournamentMissions;
    private javax.swing.JTextField tfGroupFilter;
    // End of variables declaration//GEN-END:variables
}
