package codemaptournament;

import java.awt.Color;
import java.awt.Component;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import javax.swing.DefaultCellEditor;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JTable;
import javax.swing.SwingConstants;
import javax.swing.TransferHandler;
import javax.swing.UIManager;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableColumnModel;

public class MissionTableModel extends DefaultTableModel {
    public static final DisplayAndValue[] sizes = new DisplayAndValue[] {
        new DisplayAndValue(Mission.sizeStrings[0], 0),
        new DisplayAndValue(Mission.sizeStrings[1], 1),
        new DisplayAndValue(Mission.sizeStrings[2], 2),
    };
    public static final DisplayAndValue[] complexities = new DisplayAndValue[] {
        new DisplayAndValue(Mission.complexityStrings[0], 0),
        new DisplayAndValue(Mission.complexityStrings[1], 1),
        new DisplayAndValue(Mission.complexityStrings[2], 2),
    };
    public static final int COL_INDEX = 0;
    public static final int COL_CODE = 1;
    public static final int COL_SIZE = 2;
    public static final int COL_COMPLEXITY = 3;
    public static final int NUM_COLS = 4;
    public static final int NUM_ROWS = 5;

    public MissionTableModel() {
        super(new Object[][] {
                    {new Integer(1), Mission.DEFAULT_CODE, sizes[Mission.DEFAULT_SIZE], complexities[Mission.DEFAULT_COMPLEXITY]},
                    {new Integer(2), Mission.DEFAULT_CODE, sizes[Mission.DEFAULT_SIZE], complexities[Mission.DEFAULT_COMPLEXITY]},
                    {new Integer(3), Mission.DEFAULT_CODE, sizes[Mission.DEFAULT_SIZE], complexities[Mission.DEFAULT_COMPLEXITY]},
                    {new Integer(4), Mission.DEFAULT_CODE, sizes[Mission.DEFAULT_SIZE], complexities[Mission.DEFAULT_COMPLEXITY]},
                    {new Integer(5), Mission.DEFAULT_CODE, sizes[Mission.DEFAULT_SIZE], complexities[Mission.DEFAULT_COMPLEXITY]},
            },
            new Object[] {"#", "Code", "Size", "Complexity"} );
    }
    Class[] types = new Class[]{java.lang.Integer.class, java.lang.String.class, java.lang.Object.class, java.lang.Object.class};
    boolean[] canEdit = new boolean[]{false, true, true, true};

    @Override
    public Class getColumnClass(int columnIndex) {
        return types[columnIndex];
    }

    @Override
    public boolean isCellEditable(int rowIndex, int columnIndex) {
        return canEdit[columnIndex];
    }

    @Override
    public void setValueAt(Object aValue, int row, int column) {
        if (column == COL_SIZE && !(aValue instanceof DisplayAndValue)) {
            aValue = getSizeFromString(aValue.toString());
        }
        if (column == COL_COMPLEXITY && !(aValue instanceof DisplayAndValue)) {
            aValue = getComplexityFromString(aValue.toString());
        }
        super.setValueAt(aValue, row, column);
    }
    private Object getSizeFromString(String str) {
        for (DisplayAndValue size : sizes) {
            String sizeStr = (String) size.display;
            if (sizeStr.equalsIgnoreCase(str)) {
                return size;
            }
        }
        System.err.println("unknown size " + str);
        return sizes[0];
    }
    private Object getComplexityFromString(String str) {
        for (DisplayAndValue complexity : complexities) {
            String complStr = (String) complexity.display;
            if (complStr.equalsIgnoreCase(str)) {
                return complexity;
            }
        }
        System.err.println("unknown complexity " + str);
        return complexities[0];
    }

    /**
     * Get the data in a row as a Mission object
     * @param rowIndex
     * @return
     */
    public Mission getMission(int rowIndex) {
        String code = (String) getValueAt(rowIndex, COL_CODE);
        DisplayAndValue size = (DisplayAndValue) getValueAt(rowIndex, COL_SIZE);
        DisplayAndValue complexity = (DisplayAndValue) getValueAt(rowIndex, COL_COMPLEXITY);
        return new Mission(code, (Integer) size.value, (Integer) complexity.value);
    }

    /**
     * Set model values from the given mission
     * @param row
     * @param mission
     */
    void setMission(int row, Mission mission) {
        setValueAt(mission.code, row, COL_CODE);
        setValueAt(sizes[mission.size], row, COL_SIZE);
        setValueAt(complexities[mission.complexity], row, COL_COMPLEXITY);
    }

    /**
     * Get data for all NUM_ROWS (5) rows as a list of Mission
     * @return
     */
    public List<Mission> getMissionList() {
        List<Mission> missions = new ArrayList<Mission>();
        for (int row = 0; row < NUM_ROWS; row++)
            missions.add(getMission(row));
        return missions;
    }

    /**
     * Set data for all NUM_ROWS rows using a list of Missions.
     * If the list contains less than NUM_ROWS elements, Mission.DEFAULT_MISSION
     * is used for the rest.
     * @param missions
     */
    public void setMissionList(List<Mission> missions) {
        for (int row = 0; row < NUM_ROWS; row++) {
            Mission m = row < missions.size() ? missions.get(row) : Mission.DEFAULT_MISSION;
            setMission(row, m);
        }
    }

    /**
     * Make table look pretty.
     * GUI stuff, doesn't really belong with model, but all table stuff is together.
     * Sets column widths, alignment, editors.
     * @param table
     */
    static void setUpMissionTableColumns(final JTable table) {
        TableColumnModel columnModel = table.getColumnModel();
        // set up column widths (obtained by moving it until it looked right then printing the widths)
        final int[] WIDTHS = new int[] {18, 260, 85, 85};
        for (int i = 0; i < WIDTHS.length; i++) {
            columnModel.getColumn(i).setPreferredWidth(WIDTHS[i]);
        }
        // center the number in the first column - the index
        final DefaultTableCellRenderer centeredRenderer = new DefaultTableCellRenderer();
        centeredRenderer.setHorizontalAlignment(SwingConstants.CENTER);
        columnModel.getColumn(MissionTableModel.COL_INDEX).setCellRenderer(centeredRenderer);
        // use a combo box to choose among sizes in the size column
        final DefaultCellEditor sizeEditor = new DefaultCellEditor(new JComboBox(sizes));
        columnModel.getColumn(MissionTableModel.COL_SIZE).setCellEditor(sizeEditor);
        // use a combo box to choose among complexities in the complexity column
        final DefaultCellEditor complexityEditor = new DefaultCellEditor(new JComboBox(complexities));
        columnModel.getColumn(MissionTableModel.COL_COMPLEXITY).setCellEditor(complexityEditor);

        final MissionTableModel mtm = (MissionTableModel) table.getModel();
        columnModel.getColumn(MissionTableModel.COL_SIZE).setCellRenderer(mtm.new HighlightRepeatsRenderer());
        columnModel.getColumn(MissionTableModel.COL_COMPLEXITY).setCellRenderer(mtm.new HighlightRepeatsRenderer());

        // when the model data changes for any reason, set it up to recalculate
        // which missions are repeats.
        mtm.addTableModelListener(new TableModelListener() {
            public void tableChanged(TableModelEvent e) {
                mtm.syncRepeatedMissionCheck();
                mtm.fireRepeateMissionChange();
                table.repaint(); // HACK
            }
        });
    }

    /** calculate which missions have repeated types.
     * @param missions list of missions
     * @return bitset for each mission: true = a duplicate, false = unique
     */
    public static BitSet getRepeatedMissionTypes(List<Mission> missions) {
        Map<String, BitSet> missionsWithSameType = new TreeMap<String, BitSet>();
        for (int i= 0; i < missions.size(); i++) {
            Mission mission = missions.get(i);
            final String size = Mission.sizeStrings[mission.size];
            final String complexity = Mission.complexityStrings[mission.complexity];
            final String type = size + complexity;
            if (!missionsWithSameType.containsKey(type)) {
                missionsWithSameType.put(type, new BitSet());
            }
            missionsWithSameType.get(type).set(i);
        }
        BitSet doesMissionHaveRepeatedType = new BitSet();
        for (BitSet missionsWithThatType : missionsWithSameType.values()) {
            // if there's more than one of this type, add all missions with
            // this type to the list of missions that have repeated types
            if (missionsWithThatType.cardinality() > 1)
                doesMissionHaveRepeatedType.or(missionsWithThatType);
        }
        return doesMissionHaveRepeatedType;
    }

    /*** a RepeatedMissionListener is an ActionListener that gets updated whenever
     * the repeated type status of any row might have changed.
     */
    private List<ActionListener> repeatedMissionsListeners = new ArrayList<ActionListener>();
    public void addRepeatedMissionListener(ActionListener a) {
        repeatedMissionsListeners.add(a);
    }
    private void fireRepeateMissionChange() {
        for (ActionListener listener : repeatedMissionsListeners) {
            listener.actionPerformed(null);
        }
    }

    // record which rows have repeated types
    private BitSet repeatedRows = new BitSet();
    private void syncRepeatedMissionCheck() {
        repeatedRows = getRepeatedMissionTypes(getMissionList());
    }
    public boolean isRowBad(int row) {
        return repeatedRows.get(row);
    }
    public boolean anyRowsBad() {
        return !repeatedRows.isEmpty();
    }
    /** use isRowBad() to render cells with either black text (normal) or red text (bad) */
    private class HighlightRepeatsRenderer extends DefaultTableCellRenderer {
        @Override
        public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
            Component c = super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);
            boolean isRepeated = isRowBad(row);
            Color normal = UIManager.getColor("Table.foreground");
            if (normal == null) normal = Color.black;
            Color highlighted = Color.red;
            c.setForeground(isRepeated ? highlighted : normal);
            return c;
        }

    }
    
    /**
     * Set up the drag/drop handler for the table to import and export from plain
     * tables or HTML tables.
     * @param table
     */
    static void setUpDropHandler(final JTable table) {
        table.setTransferHandler(new TransferHandler() {

            @Override
            public int getSourceActions(JComponent c) {
                return TransferHandler.COPY;
            }

            @Override
            protected Transferable createTransferable(JComponent c) {
                if (c != table) throw new RuntimeException();
                StringBuilder sb = new StringBuilder();
                int i = 0;
                for (int row : table.getSelectedRows()) {
                    if (i++ > 0) sb.append("\n");
                    int j = 0;
                    for (int col : table.getSelectedColumns()) {
                        if (j++ > 0) sb.append("\t");
                        sb.append(table.getValueAt(row, col));
                    }
                }
//                System.out.println("got " + sb.toString());
                return new StringSelection(sb.toString());
            }


            @Override
            public boolean canImport(TransferSupport support) {
                return support.isDataFlavorSupported(DataFlavor.stringFlavor);
            }

            @Override
            public boolean importData(TransferSupport support) {
                if (!canImport(support)) return false;

                int dropRow = -1, dropCol = -1;
                final Transferable transferable = support.getTransferable();

                if (support.isDrop()) { // a drop
                    JTable.DropLocation dl = (JTable.DropLocation) support.getDropLocation();
                    dropRow = dl.getRow();
                    dropCol = dl.getColumn();
                } else { // a paste (probably)
                    dropRow = table.getSelectedRow();
                    dropCol = table.getSelectedColumn();
                }
                if (dropRow < 0 || dropCol < 0) return false;
//                System.out.println("got row = " + dropRow + " col = " + dropCol);

                String[][] tabledata = ClipboardUtil.extractTableFromTransferable(transferable);
                final int tableRows = tabledata.length;
//                System.out.println("table rows = " + tableRows);

                // start at the row of the drop
                // go through each row in the tabledata until it or the destination table runs out
                int rowIndex = dropRow;
                for (int r = 0; r < tableRows && rowIndex < NUM_ROWS; r++) {
                    int cols = tabledata[r].length;
                    // if first cell starts with "Map Name", skip first row
                    if (r == 0 && cols > 0 && tabledata[r][0].startsWith("Map Name")) {
                        continue;
                    }

                    // start at the column of the drop
                    // check each column of the table, putting the next available column in the tabledata row if valid
                    int colIndex = dropCol;
                    for (int c = 0; c < cols && colIndex < NUM_COLS; colIndex++) {
                        // skip over whichever column is mapped to the index column
                        if (table.convertColumnIndexToModel(colIndex) == COL_INDEX)
                            continue;
//                        System.out.println("setting table["+rowIndex+"]["+colIndex+"] to " + tabledata[r][c]);
                        table.setValueAt(tabledata[r][c], rowIndex, colIndex);
                        // done reading from this column in the source, go to next
                        c++;
                    }
                    
                    // done filling this row in the table go to the next
                    rowIndex++;
                }

                return true;
            }
        });
    }
}
