package codemaptournament;

import codemaptournament.SortUtil.SortSpec;
import java.util.List;
import java.util.Map;
import javax.swing.JTable;

/**
 *
 * @author Paul
 */
public class Reports {

    /** optional header to the tournament map list */
    public static final String TOURNAMENT_PREAMBLE =
            "INTRO GOES HERE\n\n"
            + "[b]A few useful links:[/b]\n"
            + "[url=http://knucklecracker.com/forums/index.php?topic=7410.0]All Tournaments[/url]\n"
            + "[url=http://knucklecracker.com/creeperworld2/viewscores.php]Code Map Scoreboard[/url]\n"
            + "[url=http://knucklecracker.com/forums/index.php?topic=7477.0]Tournament Score Tool[/url]\n"
            + "\n"
            + "[b]Theme:[/b]\n"
            + "THEME GOES HERE\n"
            + "\n";

    /**
     * Generate the high score list in a forum-formatted table.
     * Format based on part by what UpperKEES posted in Chronom high scores.
     * @param highScores score data
     * @param scoreTimestampStr record when the scores were recorded
     * @return string suitable for pasting into a forum post
     */
    public static String makeForumHighScoreList(List<HighScoreEntry> highScores, String scoreTimestampStr, boolean boldTop3) {
        StringBuilder report = new StringBuilder();

        report.append(BBCode.size(BBCode.bold("Tournament Results"), "11pt"));
        report.append("\n\n");

        final int numHighScores = highScores.size();
        final int cols = 3; // rank, player, score
        final int rows = numHighScores + 1; // header + 0 hr rows
        String[][] table = new String[rows][cols];
        int col;

        // header
        int row = 0;
        col = 0;
        table[row][col++] = BBCode.bold("Rank");
        table[row][col++] = BBCode.bold("Player");
        table[row][col++] = BBCode.bold("Total Score");
        row++;

        // high score table
        for (int i = 0; i < numHighScores; i++) {
            HighScoreEntry entry = highScores.get(i);
            String rank = Integer.toString(i + 1) + ".";
            String player = entry.user;
            String score = Integer.toString(entry.totalScore);
            if (!entry.playedAllGames) {
                score += "*";
            }
            if (i < BBCode.VICTOR_COLORS.length) {
                // special colors for first 3 places
                String color = BBCode.VICTOR_COLORS[i];
                rank = BBCode.color(rank, color);
                player = BBCode.color(player, color);
                score = BBCode.color(score, color);
                // Optionally bold anything with a color
                if (boldTop3) {
                    rank = BBCode.bold(rank);
                    player = BBCode.bold(player);
                    score = BBCode.bold(score);
                }
            }
            col = 0;
            table[row][col++] = rank;
            table[row][col++] = player;
            table[row][col++] = score;
            row++;
        }
        report.append(BBCode.table(table));

        // footer
        report.append(BBCode.italic("* Did not finish all maps."));
        report.append("\n\n");
        report.append("Scores recorded at ").append(scoreTimestampStr).append(" forum time.");

        return report.toString();
    }

    /**
     * Generate a list of the missions in the tournament (suitable for posting
     * to start the tournament).
     * @param missionTableModel list of mission
     * @return string suitable for pasting into a forum post
     */
    public static String makeForumTournamentList(List<Mission> missions) {
        StringBuilder report = new StringBuilder();

        final int numMissions = missions.size();
        final int cols = 3; // for map name, size, complexity
        final int rows = numMissions + 1; // 5 missions + 1 header
        String[][] table = new String[rows][cols];
        int col;

        // the header
        int row = 0;
        col = 0;
        table[row][col++] = BBCode.color(BBCode.bold("Map Name:"), BBCode.COLOR_MAPNAME);
        table[row][col++] = BBCode.color(BBCode.bold("Size:"), BBCode.COLOR_SIZE);
        table[row][col++] = BBCode.color(BBCode.bold("Complexity:"), BBCode.COLOR_COMPLEXITY);
        row++;

        // a row for each mission
        for (Mission mission : missions) {
            col = 0;
            table[row][col++] = BBCode.COLOR_MAP_NAMES ? BBCode.color(mission.code, BBCode.COLOR_MAPNAME) : mission.code;
            table[row][col++] = BBCode.color(Mission.sizeStrings[mission.size], BBCode.COLOR_SIZE);
            table[row][col++] = BBCode.color(Mission.complexityStrings[mission.complexity], BBCode.COLOR_COMPLEXITY);
            row++;
        }

        // report is just the table no fancy stuff
        report.append(BBCode.table(table));
        return report.toString();
    }
    /**
     * Generate a list of the missions in the tournament (suitable for posting
     * to start the tournament). This has more columns of data: Gravity, Build, and Experimentals
     * @param missionTableModel list of mission
     * @return string suitable for pasting into a forum post
     */
    public static String makeForumTournamentListExtended(List<Mission> missions) {
        StringBuilder report = new StringBuilder();

        final int numMissions = missions.size();
        final int cols = 6; // for map name, size, complexity, gravity, build dir, experimentals
        final int rows = numMissions + 1; // 5 missions + 1 header
        String[][] table = new String[rows][cols];
        int col;

        // the header
        int row = 0;
        col = 0;
        table[row][col++] = BBCode.color(BBCode.bold("Map Name:"), BBCode.COLOR_MAPNAME);
        table[row][col++] = BBCode.color(BBCode.bold("Size:"), BBCode.COLOR_SIZE);
        table[row][col++] = BBCode.color(BBCode.bold("Complexity:"), BBCode.COLOR_COMPLEXITY);
        table[row][col++] = BBCode.color(BBCode.bold("Gravity:"), BBCode.COLOR_GRAVITYDIR);
        table[row][col++] = BBCode.color(BBCode.bold("Build:"), BBCode.COLOR_BUILDDIR);
        table[row][col++] = BBCode.color(BBCode.bold("Exp:"), BBCode.COLOR_HASEXPERIMENTALS);
        row++;

        // a row for each mission
        for (Mission mission : missions) {
            col = 0;
            table[row][col++] = BBCode.COLOR_MAP_NAMES ? BBCode.color(mission.code, BBCode.COLOR_MAPNAME) : mission.code;
            table[row][col++] = BBCode.color(Mission.sizeStrings[mission.size], BBCode.COLOR_SIZE);
            table[row][col++] = BBCode.color(Mission.complexityStrings[mission.complexity], BBCode.COLOR_COMPLEXITY);
            table[row][col++] = BBCode.color("DIRECTION", BBCode.COLOR_GRAVITYDIR);
            table[row][col++] = BBCode.color("DIRECTION", BBCode.COLOR_BUILDDIR);
            table[row][col++] = BBCode.color("YES/NO", BBCode.COLOR_HASEXPERIMENTALS);
            row++;
        }

        // report is just the table no fancy stuff
        report.append(BBCode.table(table));
        return report.toString();
    }


    /**
     * Get a personal scorecard. Not used anymore since we stopped posting
     * individual results on the forum. But still works
     * @param user player name
     * @param cachedMissions list of missions
     * @param scoresByGame player scores for each game
     * @param scoreTotals precalculated totals
     * @return string suitable for pasting into a forum post
     */
    public static String makeForumPersonalResults(String user, List<Mission> cachedMissions, Map<Integer, Record> scoresByGame, ScoreTotals scoreTotals) {
        StringBuilder report = new StringBuilder();
        report.append("Scores for ").append(user).append("\n");

        final int numMissions = cachedMissions.size();
        final int cols = 5; // game, size, complexity, score, time
        final int rows = numMissions + 2; // header, data, totals
        String[][] table = new String[rows][cols];
        int col;

        // header
        int row = 0;
        col = 0;
        table[row][col++] = BBCode.color(BBCode.bold("Map Name:"), BBCode.COLOR_MAPNAME);
        table[row][col++] = BBCode.color(BBCode.bold("Size:"), BBCode.COLOR_SIZE);
        table[row][col++] = BBCode.color(BBCode.bold("Complexity:"), BBCode.COLOR_COMPLEXITY);
        table[row][col++] = BBCode.bold("Score:");
        table[row][col++] = BBCode.bold("Time:");
        row++;
        // scores for each mission
        for (int i = 0; i < numMissions; i++) {
            Mission mission = cachedMissions.get(i);
            final Record record = scoresByGame.get(i);
            Integer score = (record == null) ? 0 : record.score; // 0 for empty, not blank
            Frames time = (record == null) ? new Frames(0) : record.time; // 0 for empty, not blank
            col = 0;
            table[row][col++] = BBCode.color(mission.code, BBCode.COLOR_MAPNAME);
            table[row][col++] = BBCode.color(Mission.sizeStrings[mission.size], BBCode.COLOR_SIZE);
            table[row][col++] = BBCode.color(Mission.complexityStrings[mission.complexity], BBCode.COLOR_COMPLEXITY);
            table[row][col++] = Integer.toString(score);
            String timeStr = time.toString();
            if (BBCode.RIGHT_JUSTIFY_PERSONAL_TIMES) timeStr = BBCode.right(timeStr);
            table[row][col++] = timeStr;
            row++;
        }
        // totals
        col = 0;
        table[row][col++] = BBCode.bold("Total:");
        table[row][col++] = "";
        table[row][col++] = "";
        table[row][col++] = Integer.toString(scoreTotals.getTotalScore());
        String totalTimeStr = scoreTotals.getTotalTimeAsFrames().toString();
        if (BBCode.RIGHT_JUSTIFY_PERSONAL_TIMES) totalTimeStr = BBCode.right(totalTimeStr);
        table[row][col++] = totalTimeStr;

        report.append(BBCode.table(table));
        return report.toString();
    }
    // how the columns of the player results tables should be sorted for best 3
    // the order is relative to the table model - the JTable (view) might
    // have these columns in a different order.
    private final static SortSpec [] modelSortSpecs = {
        SortSpec.DONT_CARE, // col 0 - title
        SortSpec.HI_FIRST, // col 1 - map 1
        SortSpec.HI_FIRST, // col 2 - map 2
        SortSpec.HI_FIRST, // col 3 - map 3
        SortSpec.HI_FIRST, // col 4 - map 4
        SortSpec.HI_FIRST, // col 5 - map 5
        SortSpec.HI_FIRST, // col 6 - total scores
        SortSpec.LO_FIRST, // col 7 - total times
    };
    public static String makeForumDetailedResults(JTable tablePlayerResults, List<Mission> missions, String scoreTimestampStr, boolean colorMapTop3, boolean boldTop3) {
        boolean mapNames = true; // don't know any case where we'd not want this
        StringBuilder report = new StringBuilder();
        report.append(BBCode.size(BBCode.bold("Results Per Player"), "11pt"));
        report.append("\n");

        final int srcCols = tablePlayerResults.getColumnCount();
        final int srcRows = tablePlayerResults.getRowCount();
        final int dstCols = srcCols + 1; // for rank
        final int dstRows = srcRows + 1; // for header
        int row = 0;
        int col;
        String[][] table = new String[dstRows][dstCols];
        final int COL_RANK = 0;
        // add the column headers
        for (col = 0; col < dstCols; col++) {
            String header;
            // First columns header is Rank
            if (col == COL_RANK) header = "Rank";
            // Other columns are Map1-5 (converted to mission names) or total score/time
            else {
                int table_col = col > COL_RANK ? col-1 : col;
                header = tablePlayerResults.getColumnName(table_col);
                int model_col = tablePlayerResults.convertColumnIndexToModel(table_col);
                final int MODEL_COL_MAP1 = 1, MODEL_COL_MAP5 = 5;
                if (model_col >= MODEL_COL_MAP1 && model_col <= MODEL_COL_MAP5) {
                    // maybe replace this column's boring name 'Score 1' with 'Exciting Code Thing Prime'
                    if (mapNames)
                        header = missions.get(model_col-MODEL_COL_MAP1).code;
                }
            }
            // Headers are in bold
            table[row][col] = BBCode.bold(header);
        }
        row++;

        // prepare to find the best scores for each map (and top total scores/times)
        // Need to convert the columns in the model to the possibly different
        // columns in the view (since columns can be reordered).
        SortSpec[] tableSortSpecs = new SortSpec[srcCols];
        for (int modelCol = 0; modelCol < modelSortSpecs.length; modelCol++) {
            SortSpec sortSpec = modelSortSpecs[modelCol];
            int viewCol = tablePlayerResults.convertColumnIndexToView(modelCol);
            tableSortSpecs[viewCol] = sortSpec;
        }
        // Sort the table by columns
        int[][] rankings = SortUtil.getSortIndices(tablePlayerResults, tableSortSpecs);

        // add the data from the player results table
        // complications: add in the rank column
        // format some fancily (gold/silver/bronze rows for victors, bold for column max)
        for (int sr = 0; sr < srcRows; sr++) {
            int rank = sr + 1;
            col = 0;
            for (int dc = 0; dc < dstCols; dc++) {
                String cell;
                String color = "";
                boolean colorize = false;
                // The rank column has first, second, third place as gold, silver, bronze.
                if (dc == COL_RANK) {
                    cell = Integer.toString(rank);
                    colorize = true;
                    color = BBCode.getVictorColor(sr);
                } else {
                    // find the source column (which might be off from the rendered dst column)
                    int sc = dc > COL_RANK ? dc-1 : dc;
                    final Object value = tablePlayerResults.getValueAt(sr, sc);
                    if (value == null) {
                        cell = "";
                    } else {
                        cell = value.toString();
                        
                        // Based on the rank of this cell in its column, maybe apply other formatting to cell.
                        final int rankInThisCol = rankings[sr][sc];
                        if (colorMapTop3) {
                            // in this scheme, the top 3 in each sorted column are colorized
                            // (they might also get bolded below if boldTop3 is set)
                            if (tableSortSpecs[sc] != SortSpec.DONT_CARE) {
                                colorize = true;
                                color = BBCode.getVictorColor(rankInThisCol);
                            }
                        } else {
                            // In this scheme, the only special formatting for sorted columns
                            // is that the top cell gets bolded.
                            if (rankInThisCol == 0)
                                cell = BBCode.bold(cell);
                        }
                        if (sc == 0) { // check if Name column
                            // color the names of the top 3 people, not just their ranks
                            colorize = true;
                            color = BBCode.getVictorColor(sr);
                        }
                    }
                }
                // check if this cell will have a color
                if (colorize && !color.isEmpty()) {
                    cell = BBCode.color(cell, color);
                    if (boldTop3) cell = BBCode.bold(cell); // If set, make all ranked cells bold.
                }
                // record cell in table
                table[row][col++] = cell;
            }
            row++;
        }

        report.append(BBCode.table(table));

        if (!colorMapTop3)
            report.append(BBCode.italic("Best scores in Bold.")).append("\n");
        // no helper text needed for gold/silver/bronze

        report.append("Scores recorded at ").append(scoreTimestampStr).append(" forum time.");

        return report.toString();
    }

}
