package codemaptournament;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.swing.table.AbstractTableModel;
import org.pojava.datetime.CalendarUnit;

import org.pojava.datetime.DateTime;

/**
 *
 * @author Paul
 */
public class AutoWeeklyMissionFetcher {

    public static final int MAX_MISSIONS = 5;
    public static final URL urlCodeMapBoard;

    static {
        try {
            urlCodeMapBoard = new URL("http://knucklecracker.com/forums/index.php?board=16.0");
// could use advanced search, but needs more processing http://knucklecracker.com/forums/index.php?action=search&advanced=1&search=weekly+5+map+tournament&searchtype=1&userspec=*&minage=0&maxage=9999&subject_only=1&sort=ID_MSG%7Cdesc&brd%5B16%5D=16&submit=Search
        } catch (MalformedURLException ex) {
            throw new RuntimeException(ex);
        }
    }
    // Extract threads listed on a KC forum page
    final Pattern pThreads = Pattern.compile("<span[^>]*id=\"msg_[^>]*>(.*?)</span>", Pattern.DOTALL);
    // Find links in HTML (pick out the link to the thread page from the thread title)
    final Pattern pLink = Pattern.compile("<a[^>]*href=\"([^\"]+)\"[^>]*>(.*?)</a>", Pattern.DOTALL);
    // From a thread title, match and extract the name of the tournament and its closing date
    // Must have Weekly 5 Map Tournament|Challenge, then the tourney name, then Closing or Ending [on], then the date (the remainder of the title)
    // Fields separated by hyphens or spaces
    final Pattern pWeeklyTournament = Pattern.compile("\\bWeekly\\s+(?:5|Five).*Map\\s+(?:Tournament|Challenge)[\\s-]+(.*?)[\\s-]+(?:Clos\\w*|End\\w*)\\s*(?:on\\s*)?\\b(.*)\\s*$", Pattern.CASE_INSENSITIVE);
    // In a thread, extract the individual posts
    final Pattern pPost = Pattern.compile("<div[^>]*class=\"post\"[^>]*>(.*?)</div>", Pattern.DOTALL);

    public Map<WeekInfo, List<Mission>> getWeeklyMissionData() throws IOException {
        List<WeekInfo> weekInfos = getWeeklyMissionList();
        // note - a TreeMap will sort by keys(WeekInfo)
        // WeekInfo's are comparable by closing date (descending order)
        Map<WeekInfo, List<Mission>> data = new TreeMap<WeekInfo, List<Mission>>();
        for (WeekInfo weekInfo : weekInfos) {
            List<Mission> missions = getMissionsForWeek(weekInfo);
            data.put(weekInfo, missions);
        }
        return data;
    }

    /**
     * From a tournament's url get the first post, then get a list of missions in a table
     * @param weekInfo
     * @return
     */
    private List<Mission> getMissionsForWeek(WeekInfo weekInfo) throws IOException {
        List<Mission> missions = new ArrayList<Mission>();
        String thread = NetUtil.unescape(NetUtil.readCompletely(weekInfo.link));
        Matcher mPost = pPost.matcher(thread);
        try {
            // Look at each post (on the first page) in turn...
            while (mPost.find()) {
                String postContents = mPost.group(1);
                String[][] table = ClipboardUtil.extractTableFromHTMLString(postContents);
                if (table == null) continue; // No table in this post, try the next post.

                // Look at the contents of the table, try to extract mission information
                for (String[] row : table) {
                    if (row.length < 0) {
                        continue;
                    }
                    // skip any header row (assume it the first header is map name ...)
                    if (row[0].toLowerCase().startsWith("map name")) {
                        continue;
                    }
                    // Parse all subsequent rows of data as map name first, then data
                    final String mapName = row[0];
    //                System.out.println("got " + mapName);
                    final int otherStuff = row.length - 1;
                    int size = Mission.DEFAULT_SIZE;
                    int complexity = Mission.DEFAULT_COMPLEXITY;
                    if (otherStuff >= 2) {
                        // assume next two are size and complexity
                        String sSize = row[1];
                        String sComp = row[2];
                        size = Mission.getSizeFromString(sSize);
                        complexity = Mission.getComplexityFromString(sComp);
                    }
                    Mission m = new Mission(mapName, size, complexity);
                    missions.add(m);
                    if (missions.size() >= MAX_MISSIONS) {
                        break;
                    }
                }

                // No problems so far, accept the data from the table in this post
                break;
            }
        } catch (NullPointerException npe) {
            // usually due to table being null
            Logger.getLogger(AutoWeeklyMissionFetcher.class.getName()).log(Level.WARNING, "no table found for {0}", weekInfo.link);
        }
        return missions;
    }

    /**
     * Check the Code Map Discussion board for weekly tournament threads
     * @return
     */
    private List<WeekInfo> getWeeklyMissionList() throws IOException {
        List<WeekInfo> infos = new ArrayList<WeekInfo>();
        String codeBoard = NetUtil.unescape(NetUtil.readCompletely(urlCodeMapBoard));
        Matcher mThreads = pThreads.matcher(codeBoard);
        while (mThreads.find()) {
            String msg = mThreads.group(1);
            Matcher mLink = pLink.matcher(msg);
            if (mLink.find()) {
                String link = mLink.group(1);
                String text = mLink.group(2);
                Matcher mWeeklyTournament = pWeeklyTournament.matcher(text);
                if (mWeeklyTournament.find()) { // find not matches - for now accept&ignore fluff before and after
                    String topic = mWeeklyTournament.group(1);
                    String closedate_str = mWeeklyTournament.group(2);
//                    System.out.println("Topic " + topic + " Closedate " + closedate_str);
                    // parse the close date string (note: current year is assumed if unspecified)
                    DateTime closedate = new DateTime(closedate_str);
                    final int SUNDAY = 1;
                    // the cutoff date is midnight (0:00) on the Monday of the tournament
                    // but sometimes the closing date is listed as the Sunday
                    DateTime cutoff = closedate;
                    if (closedate.weekday() == SUNDAY) {
                        // make the cutoff be at midnight on Monday, not midnight on Sunday
                        cutoff = closedate.shift(CalendarUnit.DAY, 1);
                    }

                    // note that local (midnight in forum time) is not always midnight in local time
//                    System.out.println("Parsed " + closedate_str + " as " + closedate + " cutoff= " + cutoff);
                    WeekInfo weekInfo = new WeekInfo(text, topic, cutoff.toDate(), closedate_str, new URL(link));
                    infos.add(weekInfo);
                }
            }
        }
        return infos;
    }
}

class WeekInfo implements Comparable<WeekInfo> {

    /** raw text of the thread's title */
    String fulltext;
    /** subject of the tournament */
    String topic;
    /**When the tournament ends.
     * If now is before cutoff, the tournament is still open.
     * Usually this is 0000 on Monday (right after 2359 on Sunday)
     */
    Date cutoff;
    /** human-readable closing date (what was given in the forum post) */
    String closing;
    /** link to the thread's url */
    URL link;

    public WeekInfo(String fulltext, String topic, Date cutoff, String closing, URL link) {
        this.fulltext = fulltext;
        this.topic = topic;
        this.cutoff = cutoff;
        this.closing = closing;
        this.link = link;
    }

    /** sort weeks from recent to old */
    public int compareTo(WeekInfo o) {
        return -1 * cutoff.compareTo(o.cutoff);
        // this it the same as sorting by closing date, but I store the closing date as text
        // and the cutoff date (+1day) as a Date
    }

    /** compare cutoff time to current time */
    public boolean isOpen() {
        Date now = new Date();
        return now.before(cutoff);
    }
}

class SimpleMissionTableModel extends AbstractTableModel {

    private List<Mission> missionList;

    public void setMissionList(List<Mission> missionList) {
        this.missionList = missionList;
        fireTableDataChanged();
    }

    public int getRowCount() {
        if (missionList == null) {
            return 5;
        }
        return missionList.size();
    }

    public int getColumnCount() {
        return 3;
    }

    @Override
    public Class<?> getColumnClass(int columnIndex) {
        return String.class;
    }

    public Object getValueAt(int rowIndex, int columnIndex) {
        if (missionList == null) {
            return null;
        }
        final Mission m = missionList.get(rowIndex);
        switch (columnIndex) {
            case 0:
                return m.code;
            case 1:
                return Mission.sizeStrings[m.size];
            case 2:
                return Mission.complexityStrings[m.complexity];
        }
        throw new IllegalArgumentException("column index out of bounds");
    }
}
