package codemaptournament;

import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * simple utility for setting the clipboard contents to a string
 * @author Paul
 */
public class ClipboardUtil {
    public static void setClipboardContents(String contents) {
        ClipboardOwner co = new ClipboardOwner() {
            public void lostOwnership(Clipboard clipboard, Transferable contents) {}
        };
        StringSelection selection = new StringSelection(contents);
        Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
        clipboard.setContents(selection, co);
    }

    /**
     * If the transferable can be html text, look for an html-table.
     * If that fails, if it can be plain text, look for a plain table.
     *
     * @param transferable Transferable from TransferHandler
     * @return A 2D array of strings corresponding to the found table, or null if no table was found.
     */
    public static String[][] extractTableFromTransferable(Transferable transferable) {
        try {
            String htmlText = (String) transferable.getTransferData(htmlFlavor);
            String[][] ret = extractTableFromHTMLString(NetUtil.unescape(htmlText));
            if (ret != null) return ret;
        } catch (Exception e) {
//            System.err.println("can't get html table");
        }

        try {
            String plainData = (String) transferable.getTransferData(DataFlavor.stringFlavor);
            String[][] ret = extractTableFromPlainString(plainData);
            if (ret != null) return ret;
        } catch (Exception e) {
//            System.err.println("can't get plain table");
        }

        return null;
    }

    private static final DataFlavor htmlFlavor;
    static {
        try {
            htmlFlavor = new DataFlavor("text/html; class=java.lang.String; charset=UTF-8");
        } catch (ClassNotFoundException ex) {
            throw new RuntimeException(ex);
        }
    }


    // throw away html tag innards except for the tag name
    private static final Pattern pTidyTags = Pattern.compile("<(\\w+)\\b[^>]+>");
    private static final String rTidyTags = "<$1>";
    // find regions inside <tr>...</tr>
    private static final Pattern pRow = Pattern.compile("<tr[^>]*>(.*?)</tr>");
    // find regions inside <td>...</td>
    private static final Pattern pElem = Pattern.compile("<td[^>]*>(.*?)</td>");
    // throw away all html-tags
    private static final Pattern pStripTags = Pattern.compile("<[^>]+>");
    private static final String rStripTags = "";

    /**
     * Examine the string (which likely came from the clipboard / drag&drop)
     * to find an html-table.
     * @param contents a String to examine
     * @return A 2D array of strings corresponding to the found table, or null if no table was found.
     */
    public static String[][] extractTableFromHTMLString(String contents) {
        // First throw away anything inside a tag that's not the tag name
        Matcher mTidyTags = pTidyTags.matcher(contents);
        String justtags = mTidyTags.replaceAll(rTidyTags);

        // then go through each table row (tr) to each table datum (td) to the raw text (stripping tags)
        Matcher mRow = pRow.matcher(justtags);
        Matcher mElem = pElem.matcher("");
        Matcher mStripTags = pStripTags.matcher("");

        List<List<String>> rows = new ArrayList<List<String>>();
        while (mRow.find()) {
            List<String> row = new ArrayList<String>();
            final String rowdata = mRow.group(1);
            mElem.reset(rowdata);
            while (mElem.find()) {
                mStripTags.reset(mElem.group(1));
                row.add(mStripTags.replaceAll(rStripTags));
            }
            rows.add(row);
        }

        final int rowCount = rows.size();
        if (rowCount == 0) return null;

        // convert from List<String> to [String]
        String[][] ret = new String[rowCount][];
        for (int i= 0; i < rowCount; i++) {
            ret[i] = rows.get(i).toArray(new String[0]);
        }
//        System.out.println("html = [");
//        for (String[] row : ret) {
//            System.out.println(Arrays.toString(row));
//        }
//        System.out.println("]");
        return ret;
    }

    /**
     * Examine the string (which likely came from the clipboard / drag&drop)
     * to find a plain table (rows separated by enter, columns separated by tab)
     * NEW: when you can't find a table, just assume it is a 1x1 table
     * and return the contents itself
     * @param contents a String to examine
     * @return A 2D array of strings corresponding to the found table, or the string itself if no table was found.
     */
    public static String[][] extractTableFromPlainString(String contents) {
        String[] rows = contents.split("\n");
        final int rowCount = rows.length;
        if (rowCount == 0) return new String[][] {{contents}};

        String[][] tabledata = new String[rowCount][];
        for (int i = 0; i < rowCount; i++) {
            final String row = rows[i];
            String[] cols = row.split("\t");
            tabledata[i] = cols;
        }

//        System.out.println("plain string = [");
//        for (String[] row : tabledata) {
//            System.out.println(Arrays.toString(row));
//        }
//        System.out.println("]");

        return tabledata;
    }
}
