package codemaptournament;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * Feed in data and the ResultRanker spits out the high score chart.
 * @author Paul
 */
public class ResultRanker {

    private Map<String, Map<Integer, Record>> scoresByGameByUser;
    private Map<String, ScoreTotals> totalsByUser;
    private final int missionCount;

    ResultRanker(int missionCount) {
        scoresByGameByUser = new TreeMap<String, Map<Integer, Record>>(String.CASE_INSENSITIVE_ORDER);
        totalsByUser = new TreeMap<String, ScoreTotals>(String.CASE_INSENSITIVE_ORDER);
        this.missionCount = missionCount;
    }

    public Map<String, Map<Integer, Record>> getScoresByGameByUser() {
        return scoresByGameByUser;
    }

    public Map<String, ScoreTotals> getTotalsByUser() {
        return totalsByUser;
    }

    /**
     * Record one data point. Call this for every entry you wish to record.
     * @param user name of the person
     * @param totalScore that person's totalScore
     * @param gameNumber integer, 0 = first game (though not strictly necessary)
     */
    public void recordScore(int gameNumber, Record record) {
        String user = record.user;
        if (!scoresByGameByUser.containsKey(user)) {
            scoresByGameByUser.put(user, new TreeMap<Integer, Record>());
            totalsByUser.put(user, new ScoreTotals(0, 0));
        }
        scoresByGameByUser.get(user).put(gameNumber, record);
        totalsByUser.get(user).add(record.score, record.time);
    }

    /**
     * Get a list of HighScoreEntrys in order from first place to last.
     * Ordering is by # of games completed (high-low), then by totalScore(high-low),
     * then by user name (a-z).
     * @return list of high totalScore entries
     */
    public List<HighScoreEntry> getHighScoreList() {
        List<HighScoreEntry> highScores = new ArrayList<HighScoreEntry>();
        for (Map.Entry<String, Map<Integer, Record>> entry : scoresByGameByUser.entrySet()) {
            String user = entry.getKey();
            Map<Integer, Record> scoresByGame = entry.getValue();
            ScoreTotals totals = totalsByUser.get(user);
            int gamesPlayed = scoresByGame.size();
            highScores.add(new HighScoreEntry(
                    user,
                    totals.getTotalScore(),
                    totals.getTotalTimeAsFrames(),
                    gamesPlayed,
                    gamesPlayed == missionCount));
        }

        Collections.sort(highScores, HighScoreEntry.tournamentComparator);
        return highScores;
    }
}
