package codemaptournament;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.List;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

/**
 * Contact high score server to get list of high scores for a mission
 * @author Paul
 */
public class ScoreFetcher {
    /** Delay for each fetchScoreRecords request so we don't swamp the server */
    public int DELAY_MS = 100;

    // constants for the options for sorting & the time filter
    public static final String SORT_SCORES = "scores";
    public static final String SORT_TIMES = "time";
    public static final int TIME_LAST_FEW_DAYS = 1;
    public static final int TIME_LAST_WEEK = 2;
    public static final int TIME_LAST_MONTH = 3;
    public static final int TIME_ALL_TIME = 0;
    public static final DisplayAndValue[] sorts = {
        new DisplayAndValue("Top Scores", SORT_SCORES),
        new DisplayAndValue("Top Times", SORT_TIMES),
    };
    public static final String longestTimeFilterStr = "Last Few Days";
    // index in the array is the integer constant TIME_....
    public static final DisplayAndValue[] timeFilters = {
        new DisplayAndValue("All Time", TIME_ALL_TIME),
        new DisplayAndValue("Last Few Days", TIME_LAST_FEW_DAYS),
        new DisplayAndValue("Last Week", TIME_LAST_WEEK),
        new DisplayAndValue("Last Month", TIME_LAST_MONTH),
    };


    private String sort = SORT_SCORES;
    private Integer timefilter = TIME_ALL_TIME;
    private String groupfilter;
    private static DocumentBuilder builder;

    static {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setValidating(false);
        factory.setExpandEntityReferences(false);
        try {
            builder = factory.newDocumentBuilder();
        } catch (ParserConfigurationException ex) {
            throw new RuntimeException(ex);
        }
    }

    public String getSort() {
        return sort;
    }

    public void setSort(String sort) {
        this.sort = sort;
    }

    public Integer getTimefilter() {
        return timefilter;
    }

    public void setTimefilter(Integer timefilter) {
        this.timefilter = timefilter;
    }

    public String getGroupfilter() {
        return groupfilter;
    }

    public void setGroupfilter(String groupfilter) {
        this.groupfilter = groupfilter;
    }

    

    /**
     * Get a list of score records for a game.
     * Uses the group filter, time filter, and sort methods if defined.
     * Also it sleeps for DELAY_MS as a courtesy to the server. I dunno how
     * useful that is, but it seems safe.
     * @param gameUID the UID of the game
     * @return list of score records
     * @throws Exception if anything bad happens (usually an internet error)
     */
    public List<Record> fetchScoreRecords(String gameUID) throws Exception {
        String xml = fetchScores(gameUID);
        Document xmlDocument = parseXML(xml);
        List<Record> records = getRecordsFromXML(xmlDocument);
        Thread.sleep(DELAY_MS); // so we don't swamp the server
        return records;
    }

    private String makeParamString(String gameUID) {
        StringBuilder sb = new StringBuilder();
        sb.append("gameUID=");
        sb.append(gameUID);
//        sb.append('&');
//        sb.append("userfilter=");
//        sb.append(tfUserFilter.getText());
        if (groupfilter != null && !groupfilter.isEmpty()) {
            sb.append('&');
            sb.append("groupfilter=");
            sb.append(NetUtil.escapeURL(groupfilter));
        }
        if (sort != null && !sort.isEmpty()) {
            sb.append('&');
            sb.append("sort=");
            sb.append(sort); // sort never needs to be escaped
        }
        if (timefilter != null) {
            sb.append('&');
            sb.append("timefilter=");
            sb.append(timefilter); // time filter never needs to be escaped
        }
//        System.out.println("query params = " + sb);
        return sb.toString();
    }
    private String fetchScores(String gameUID) throws Exception {
        final String scoreQuery = "http://knucklecracker.com/creeperworld2/scoreQuery.php";
        URL url = new URL(scoreQuery + '?' + makeParamString(gameUID));
        URLConnection conn = url.openConnection();
        // Get the response
        BufferedReader rd = new BufferedReader(new InputStreamReader(conn.getInputStream()));
        StringBuilder sb = new StringBuilder();
        String line;
        while ((line = rd.readLine()) != null) {
            sb.append(line);
        }
        rd.close();
        return sb.toString();
    }

    private static Document parseXML(String xml) throws Exception {
        builder.reset();
        Document xmlDocument;
        xmlDocument = builder.parse(new InputSource(new StringReader(sanitize(xml))));
        return xmlDocument;
    }

    /**
     * Escape & in input so we don't get an xml parse error related to undefined entities.
     * This is because strict XML can't have any fancy entities, but knucklecracker's
     * score query sometimes returns names with special letters encoded with entities.
     * (I forget which map/player this was for) // TODO see if this ever happens again.
     * @param input
     * @return
     */
    private static String sanitize(String input) {
        return input.replaceAll("&", "&amp;");
    }

    private static List<Record> getRecordsFromXML(Document xmlDocument) {
        NodeList records = xmlDocument.getElementsByTagName("record");
        List<Record> result = new ArrayList<Record>();
        // process each record
        for (int i= 0; i < records.getLength(); i++) {
            // try to get the details for the record (store as a Record object)
            Node record = records.item(i);
            NodeList childNodes = record.getChildNodes();
            Integer rank = null, score = null, time = null, plays = null;
            String user = null;
            for (int j = 0; j < childNodes.getLength(); j++) {
                Node child = childNodes.item(j);
                //System.out.println("child=" + child + "," + child.getNodeName() + "," + child.getNodeValue());
                if (child.getNodeType() == Node.ELEMENT_NODE) {
                    String sub = child.getNodeName();
                    String value = child.getTextContent();
                    try {
                        if (!value.isEmpty()) {
                            if (Record.RANK.equals(sub)) {
                                rank = Integer.parseInt(value);
                            } else if (Record.USER.equals(sub)) {
                                user = value;
                            } else if (Record.SCORE.equals(sub)) {
                                score = Integer.parseInt(value);
                            } else if (Record.TIME.equals(sub)) {
                                time = Integer.parseInt(value);
                            } else if (Record.PLAYS.equals(sub)) {
                                plays = Integer.parseInt(value);
                            }
                        }
                    } catch (NumberFormatException nfe) {
                    }
                }
            }
            // done parsing all children of this result
            // check if we have enough data recovered to add a row to the table
            if (rank != null && score != null && time != null && plays != null && user != null) {
                Record r = new Record(rank, user, score, time, plays);
//                System.out.println(rank + "," + user + "," + score + "," + time + "," + plays);
                result.add(r);
            }
        }
        return result;
    }


}
