package codemaptournament;

import java.util.Arrays;
import java.util.Comparator;
import javax.swing.JTable;
import org.apache.commons.collections.Transformer;
import org.apache.commons.collections.comparators.NullComparator;
import org.apache.commons.collections.comparators.ReverseComparator;
import org.apache.commons.collections.comparators.TransformingComparator;

/**
 * Functions for sorting JTables (and anything rectangular, really).
 * @author Paul
 */
public class SortUtil {

    public enum SortSpec {HI_FIRST, LO_FIRST, DONT_CARE};

    // Sort some Comparables and keep track of their original indices
    // See http://stackoverflow.com/questions/951848/java-array-sort-quick-way-to-get-a-sorted-list-of-indices-of-an-array
    public static class SortPair implements Comparable<SortPair>
    {
        private int originalIndex;
        private Comparable value;

        public SortPair(Comparable value, int originalIndex) {
            this.value = value;
            this.originalIndex = originalIndex;
        }

        @Override
        public int compareTo(SortPair o) {
            return value.compareTo(o.getValue());
        }

        public int getOriginalIndex() {
            return originalIndex;
        }

        public Comparable getValue() {
            return value;
        }

        public static final Transformer valueExtractor = new Transformer() {
            public Object transform(Object input) {
                return ((SortPair) input).getValue();
            }
        };
    }

    /**
     * Determine the rank of every item in the JTable, comparing each column
     * separately. Use an array of SortSpec to determine which direction to sort
     * each column: HI_FIRST (largest value gets rank 0), LO_FIRST (smallest),
     * or DONT_CARE (all spots in this column get rank -1). All nulls are sorted
     * last.
     * @param table rectangular data of Comparable data (will not be modified)
     * @param sortSpecs how to sort each column
     * @return 2D array of indices, same dimensions [row][col] as table.
     */
    public static int[][] getSortIndices(JTable table, SortSpec[] sortSpecs) {
        final int cols = table.getColumnCount();
        final int rows = table.getRowCount();
        int[][] indices = new int[rows][cols];

        for (int col = 0; col < cols; col++) {
            SortSpec spec = sortSpecs[col];
            if (spec == SortSpec.DONT_CARE) {
                for (int row = 0; row < rows; row++)
                    indices[row][col] = -1;
            } else {
                // Figure out how to sort the data in this column
                final boolean reverse = spec == SortSpec.HI_FIRST;
                final boolean nullsAreHigh = !reverse; // this is what makes nulls always sort last

                // Compare the data by extracting the value from the Sort Pair, then comparing
                // on that (which might be null, so use a NullComparator).
                // Optionally, reverse the order of the sort.
                Comparator comparator = new TransformingComparator(SortPair.valueExtractor, new NullComparator(nullsAreHigh));
                if (reverse) comparator = new ReverseComparator(comparator);


                // Build the list of data to sort (values, and original index)
                SortPair [] data = new SortPair[rows];
                for (int row = 0; row < rows; row++) {
                    Comparable x = (Comparable) table.getValueAt(row, col);
                    data[row] = new SortPair(x, row);
                }
                // Sort the data
                Arrays.sort(data, comparator);
                // From the sorted data, record the index of each item
                for (int i = 0; i < rows; i++) {
                    int row = data[i].originalIndex;
                    indices[row][col] = i;
                }
            }
        }
        return indices;
    }


    ////////////// OLD FUNCTIONALITY to find min/max of each column of table /////////////
    public enum minmaxSearchSpec {FIND_MAX, FIND_MIN, DONT_CARE};
    /**
     * Find the rows with extreme values in a table.
     * Note: indices are relative to the view (JTable) not model.
     * @param table rectangular table with elements that are all Comparable
     * @param searchSpecs list specifying whether to find the max of the column or min or skip it
     * @return list of indices for each column (-1 if none found or dont care)
     */
    public static int[] getMaxIndices(JTable table, minmaxSearchSpec[] searchSpecs) {
        final int cols = table.getColumnCount();
        final int rows = table.getRowCount();
        int[] indices = new int[cols];
        for (int col = 0; col < cols; col++) {
            indices[col] = -1; // default value
            minmaxSearchSpec searchSpec = searchSpecs[col];
            if (searchSpec == minmaxSearchSpec.DONT_CARE) continue;
            //. start searching
            Comparable limit = null;
            int index = -1;
            for (int row = 0; row < rows; row++) {
                final Object o = table.getValueAt(row, col);
                if (o == null) continue;
                final Comparable value = (Comparable) o;
                if (limit == null) {
                    // this value is the first non-null value - use it
                    limit = value;
                    index = row;
                } else {
                    final int compare = value.compareTo(limit);
                    if (searchSpec == minmaxSearchSpec.FIND_MAX && compare > 0
                     || searchSpec == minmaxSearchSpec.FIND_MIN && compare < 0) {
                        // new extreme found - record limit and index
                        limit = value;
                        index = row;
                    }
                }
            }
            indices[col] = index;
        }
        return indices;
    }
}
